//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (string.IsNullOrEmpty(Params.DavUrlPrefix) || !Params.AddRoutingConfig)  return String.Empty; #>
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Configuration;

namespace ITHit.WebDav.Server.SampleModel.Config
{
    /// <summary>
    /// DavLocation configuration element.
    /// Specifies which url for which verb shall be handled by <see cref="DavHandler"/>.
    /// <see cref="UrlRoutingModule "/> inserts routing ignore rules for these paths and verbs.
    /// </summary>
    public sealed class DavLocation : ConfigurationElement
    {
        private static readonly ConfigurationProperty propPath = 
            new ConfigurationProperty(
                "path",
                typeof(string),
                string.Empty,
                null,
                null,
                ConfigurationPropertyOptions.IsRequired);

        private static readonly ConfigurationProperty propVerbs = 
            new ConfigurationProperty(
                "verbs",
                typeof(string),
                string.Empty,
                null,
                null,
                ConfigurationPropertyOptions.None);

        private static readonly ConfigurationPropertyCollection properties = new ConfigurationPropertyCollection();

        static DavLocation()
        {
            properties.Add(propPath);
            properties.Add(propVerbs);
        }

        /// <summary>
        /// Gets relative path all resources inside which shall be treated as WebDav requests (for
        /// which ASP.NET MVC routing shall be disabled).
        /// </summary>
        [ConfigurationProperty("path", DefaultValue = "")]
        public string Path
        {
            get
            {
                return (string)base[propPath];
            }
        }

        /// <summary>
        /// Gets comma separated list of verbs that shall be treated as webdav requests (for which
        /// ASP.NET MVC routing shall be disabled).
        /// If empty - all verbs will be handled by <see cref="DavHandler"/>.
        /// </summary>
        [ConfigurationProperty("verbs", DefaultValue = "")]
        public string Verbs
        {
            get
            {
                return (string)base[propVerbs];
            }
        }

        protected override ConfigurationPropertyCollection Properties
        {
            get
            {
                return properties;
            }
        }
    }
}
