//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Hosting != HostingEnvironment.Listener)  return string.Empty; #>
using System;
using System.Configuration;
using System.IO;
//!<# if (Params.Storage == StorageType.Db) {#>
using System.Data.SqlClient;
//!<# } #>
using System.Net;
using System.Reflection;
//!<# if (Params.ACL && Params.Storage == StorageType.Ntfs) { #>
using System.Security.Principal;
//!<# } #>
//!<# if (Params.RunAsService) { #>
using System.ServiceProcess;
//!<# } #>
using System.Text;
using System.Xml;
//!<# if (Params.DeltaV) {#>
using System.Collections.Generic;
using ITHit.WebDAV.Server.DeltaV;
//!<# } #>
using ITHit.WebDAV.Server.Logger;
using ITHit.WebDAV.Server;

namespace ITHit.WebDav.Server.SampleModel
{
    /// <summary>
    /// WebDAV engine host.
    /// </summary>
    internal class Program
    {
        public static bool Listening { get; set; }

        //!<# if (Params.AuthenticationType == AuthenticationType.Digest) { #>
        private static readonly __Authentication__.DigestAuthenticationProvider digestProvider =
            new __Authentication__.DigestAuthenticationProvider(GetPasswordAndRoles);
        //!<#}#>

        private static DavEngine engine;

        //!<# if (Params.Storage == StorageType.Ntfs) { #>
        private static readonly string repositoryPath =
            ConfigurationManager.AppSettings["RepositoryPath"].TrimEnd('\\');
        //!<#}#>

        //!<# if (Params.ACL && Params.Storage == StorageType.Ntfs) { #>
        private static readonly string domain =
            ConfigurationManager.AppSettings["Domain"];

        private static readonly string anonymousUserLogin =
            ConfigurationManager.AppSettings["AnonymousUserLogin"];

        private static readonly string anonymousUserPassword =
            ConfigurationManager.AppSettings["AnonymousUserPassword"];

        private static WindowsIdentity anonymousUser;
        //!<#}#>

        //!<# if (Params.DeltaV) { #>
        private static readonly bool autoPutUnderVersionControl =
            Convert.ToBoolean(ConfigurationManager.AppSettings["AutoPutUnderVersionControl"]);
        //!<#}#>

        /// <summary>
        /// Whether requests/responses shall be logged.
        /// </summary>
        private static readonly bool debugLoggingEnabled =
            ConfigurationManager.AppSettings["DebugLoggingEnabled"].Equals(
                "true",
                StringComparison.InvariantCultureIgnoreCase);

        /// <summary>
        /// Logger instance.
        /// </summary>
        private static readonly DefaultLoggerImpl logger = new DefaultLoggerImpl();

        //!<# if (Params.RunAsService && Params.RunAsConsole) { #>
        /// <summary>
        /// Gets a value indicating whether the program is run as Windows service or standalone application.
        /// </summary>
        private static bool IsServiceMode
        {
            get { return !Environment.UserInteractive; }
        }
        //!<#}#>

        /// <summary>
        /// Entry point.
        /// </summary>
        /// <param name="args">Command line arguments.</param>
        public static void Main(string[] args)
        {
            try
            {
                CheckConfigErrors();
                Init();

                //!<# if (Params.RunAsService && Params.RunAsConsole) { #>
                if (IsServiceMode)
                {
                //!<#}#>
                    //!<# if (Params.RunAsService) { #>
                    Listening = false;
                    ServiceBase.Run(new Service());
                    //!<#}#>
                //!<# if (Params.RunAsService && Params.RunAsConsole) { #>
                }
                else
                {
                //!<#}#>
                    //!<# if (Params.RunAsConsole) { #>
                    Listening = true;
                    ThreadProc();
                    //!<#}#>
                //!<# if (Params.RunAsService && Params.RunAsConsole) { #>
                }
                //!<#}#>
            }
            catch (Exception ex)
            {
                //!<# if (Params.RunAsService && Params.RunAsConsole) { #>
                if (IsServiceMode)
                {
                //!<#}#>
                    //!<# if (Params.RunAsService) { #>
                    string errorMessage = "Could not start service\n     " + ex.Message;
                    logger.LogError(errorMessage, null);
                    //!<#}#>
                //!<# if (Params.RunAsService && Params.RunAsConsole) { #>
                }
                else
                {
                //!<#}#>
                    //!<# if (Params.RunAsConsole) { #>
                    Console.WriteLine(ex.Message);
                    Console.WriteLine("\nPress any key...");
                    Console.ReadKey();
                    //!<#}#>
                //!<# if (Params.RunAsService && Params.RunAsConsole) { #>
                }
                //!<#}#>
            }
        }

        static void Init()
        {
            var exePath = Path.GetDirectoryName(new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath);
            string logPath = Path.Combine(Directory.GetParent(exePath).FullName, @"App_Data\WebDav\Logs");
            logger.LogFile = Path.Combine(logPath, "WebDAVlog.txt");
            logger.IsDebugEnabled = debugLoggingEnabled;

            engine = new DavEngine
            {
                Logger = logger,
                OutputXmlFormatting = logger.IsDebugEnabled ? Formatting.Indented : Formatting.None,
                //!<# if (Params.DeltaV) { #>
                AutoPutUnderVersionControl = autoPutUnderVersionControl,
                //!<# } #>
            };

            //!<# if (!string.IsNullOrEmpty(Params.License)) { #>
            //!            engine.License = @"<#=Params.License == null ? null : Params.License.Replace("\"", "\"\"")#>";
            //!<# } else { #>
            engine.License = File.ReadAllText(Path.Combine(exePath, "License.lic"));
            //!<# } #>

            //!<# if (Params.AddCustomGetHandler) { #>
            // set custom handler to process GET requests to folders and display info about how to connect to server
            MyCustomGetHandler handler = new MyCustomGetHandler();
            handler.OriginalHandler = engine.RegisterMethodHandler("GET", handler);
            //!<# } #>
        }

        public static void ThreadProc()
        {
            string uriPrefix = ConfigurationManager.AppSettings["ListenerPrefix"];
            HttpListener listener = new HttpListener();
            listener.Prefixes.Add(uriPrefix);

            //!<# if (Params.AuthenticationType == AuthenticationType.None) { #>
            listener.AuthenticationSchemes = AuthenticationSchemes.Anonymous;
            //!<#}#>
            //!<# if (Params.AuthenticationType == AuthenticationType.Basic) { #>
            listener.AuthenticationSchemes = AuthenticationSchemes.Basic;
            //!<#}#>
            //!<# if (Params.AuthenticationType == AuthenticationType.Digest) { #>
            listener.AuthenticationSchemes = AuthenticationSchemes.Anonymous;
            //!<#}#>
            //!<# if (Params.AuthenticationType == AuthenticationType.Ntlm) { #>
            listener.AuthenticationSchemes = AuthenticationSchemes.IntegratedWindowsAuthentication;
            listener.UnsafeConnectionNtlmAuthentication = true; // Required to avoid buffering on a client side in case NTLM is used. Otherwise "This request requires buffering data to succeed" exception may be thrown on a client side. Also required to support upload in web browsers when NTLM is used.
            //!<#}#>

            listener.IgnoreWriteExceptions = true;

            //!<# if (Params.RunAsService && Params.RunAsConsole) { #>
#if DEBUG
            if (!IsServiceMode)
#endif
            //!<# } #>
            //!<# if (Params.RunAsConsole) { #>
#if DEBUG
            // For the sake of the development convenience, this code opens default web browser
            // with this server url when project is started in the debug mode as a console app.
            System.Diagnostics.Process.Start(uriPrefix.Replace("+", "localhost"), null);
#endif
            //!<# } #>

            listener.Start();

            Console.WriteLine("Start listening on " + uriPrefix);
            Console.WriteLine("Press Control-C to stop listener...");

            while (Listening)
            {
                IAsyncResult result = listener.BeginGetContext(ListenerCallback, listener);
                result.AsyncWaitHandle.WaitOne();
            }
        }

        //$<Listener.DavContextBase
        public static void ListenerCallback(IAsyncResult result)
        {
            HttpListener listener = (HttpListener)result.AsyncState;
            HttpListenerContext context = listener.EndGetContext(result);
            MacOsXPreprocessor.Process(context.Request); // fixes headers for Mac OS X v10.5.3 or later

            //!<# if (Params.AuthenticationType == AuthenticationType.Basic || Params.AuthenticationType == AuthenticationType.Digest) { #>
            if (!UserAutorized(context))
            {
                context.Response.StatusCode = 401;
                ShowLoginDialog(context, context.Response);
                CloseResponse(context);
                return;
            }
            //!<#}#>

            context.Response.SendChunked = false;

            //!<# if (Params.Storage == StorageType.Ntfs) { #>
            var ntfsDavContext =
                new __NtfsStorage__.DavContext(context, listener.Prefixes, repositoryPath, engine.Logger);
            //!<# if (Params.ACL && Params.Storage == StorageType.Ntfs) { #>
            ntfsDavContext.AnonymousUser = anonymousUser;
            engine.Logger.LogDebug(
                string.Format(
                    "Logged in as:{0}, anonymous user is:{1}",
                    ntfsDavContext.User.Name,
                    anonymousUser == null ? "null" : anonymousUser.Name));
            //!<#}#>
            engine.Run(ntfsDavContext);
            //!<#}#>

            //!<# if (Params.Storage == StorageType.Db) { #>
            using (var sqlDavContext = new __SqlStorage__.DavContext(context, listener.Prefixes, context.User))
            {
                engine.Run(sqlDavContext);
            }
            //!<#}#>

            //!<# if (Params.Storage == StorageType.Blank) { #>
            var davContext = new __DavImpl__.DavContext(context, listener.Prefixes, engine.Logger);
            engine.Run(davContext);
            //!<#}#>

            //!<# if (Params.AuthenticationType != AuthenticationType.None) { #>
            if (context.Response.StatusCode == 401)
            {
                ShowLoginDialog(context, context.Response);
            }
            //!<#}#>

            CloseResponse(context);
        }
        //$>

        private static void CloseResponse(HttpListenerContext context)
        {
            try
            {
                context.Response.Close();
            }
            catch
            {
                // client closed connection before the content was sent
            }
        }
        //$>

        //!<# if (Params.AuthenticationType == AuthenticationType.Digest) { #>
        private static __Authentication__.DigestAuthenticationProvider.PasswordAndRoles GetPasswordAndRoles(
            string username)
        {
            if (username == "User1")
            {
                return new __Authentication__.DigestAuthenticationProvider.PasswordAndRoles("pwd", new[] { "admin" });
            }

            return null;
        }
        //!<#}#>

        //!<# if (Params.AuthenticationType != AuthenticationType.None) { #>
        private static void ShowLoginDialog(HttpListenerContext context, HttpListenerResponse response)
        {
            //!<# if (Params.AuthenticationType == AuthenticationType.Basic) { #>
            context.Response.AddHeader("WWW-Authenticate", "Basic Realm=\"My WebDAV Server\"");
            //!<#}#>
            //!<# if (Params.AuthenticationType == AuthenticationType.Digest) { #>
            context.Response.AddHeader("WWW-Authenticate", digestProvider.GetChallenge());
            //!<#}#>
            //!<# if (Params.AuthenticationType == AuthenticationType.Ntlm) { #>
            response.AddHeader("WWW-Authenticate", "Negotiate");
            //!<#}#>

            if (context.Request.HttpMethod.ToLower() != "options"
                && context.Request.HttpMethod.ToLower() != "head")
            {
                byte[] message = new UTF8Encoding().GetBytes("Access denied");
                context.Response.ContentLength64 = message.Length;
                context.Response.OutputStream.Write(message, 0, message.Length);
            }
        }
        //!<#}#>

        //!<# if (Params.AuthenticationType == AuthenticationType.Basic || Params.AuthenticationType == AuthenticationType.Digest) { #>
        private static bool UserAutorized(HttpListenerContext context)
        {
            //!<# if (Params.AuthenticationType == AuthenticationType.Basic) { #>
            HttpListenerBasicIdentity identity = (HttpListenerBasicIdentity) context.User.Identity;
            string username = identity.Name;

            // workaround for Windows Vista Basic authentication.
            // User name may be submitted in the following format: Machine\User.
            int ind = username.LastIndexOf('\\');
            if (ind > 0)
            {
                username = username.Remove(0, ind + 1);
            }

            return username == "User1" && identity.Password == "pwd";
            //!<#}#>
            //!<# if (Params.AuthenticationType == AuthenticationType.Digest) { #>
            // workaround for Windows Server 2003 and XP.
            // Digest Authorization header is not attched to OPTIONS requests.
            if (context.Request.HttpMethod.ToLower() == "options")
            {
                return true;
            }

            return digestProvider.AuthenticateRequest(context.Request.Headers, context.Request.HttpMethod) != null;
            //!<#}#>
        }
        //!<#}#>

        /// <summary>
        /// Checks configuration errors.
        /// </summary>
        private static void CheckConfigErrors()
        {
            //!<# if (Params.DeltaV) {#>
            string autoput = ConfigurationManager.AppSettings["AutoPutUnderVersionControl"];
            bool v;
            if (!bool.TryParse(autoput, out v))
            {
                string message = "AutoPutUnderVersionControl section is missing or invalid!" +
                    " Should contain 'true/false' value only.";
                throw new Exception(message);
            }
            //!<# } #>

            //!<# if (Params.Storage == StorageType.Db) {#>
            string connStr = ConfigurationManager.ConnectionStrings["WebDAV"].ConnectionString;
            if (string.IsNullOrEmpty(connStr))
            {
                throw new Exception("SqlConnectionString section is missing or invalid!");
            }

            var exePath = Path.GetDirectoryName(new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath);
            var dataPath = Path.Combine(Directory.GetParent(exePath).FullName, "App_Data");
            connStr = connStr.Replace("|DataDirectory|", dataPath);
            // test database connection
            using (SqlConnection conn = new SqlConnection(connStr))
            {
                conn.Open();
            }
            //!<# } #>

            //!<# if (Params.Storage == StorageType.Ntfs) {#>
            string repPath = ConfigurationManager.AppSettings["RepositoryPath"];
            if (repPath == null || !Directory.Exists(repPath))
            {
                throw new Exception("Invalid RepositoryPath configuration parameter value.");
            }
            //!<# } #>

            //!<# if (Params.DeltaV) {#>
            string autoVerMode = ConfigurationManager.AppSettings["AutoVersionMode"];
            if (autoVerMode == null)
            {
                throw new Exception("AutoVersionMode is not specified.");
            }

            if (!new List<string>(Enum.GetNames(typeof(AutoVersion))).Contains(autoVerMode))
            {
                throw new Exception(
                        "Unsupported AutoVersionMode configuration parameter value.\n" +
                        "Possible values are: " +
                        "(NoAutoVersioning | CheckOutCheckIn | CheckOutUnlockedCheckIn | CheckOut | LockedCheckOut).");
            }
            //!<# } #>

            string uriPrefix = ConfigurationManager.AppSettings["ListenerPrefix"];
            if (string.IsNullOrEmpty(uriPrefix))
            {
                throw new Exception("ListenerPrefix section is missing or invalid!");
            }
        }
    }
}
