//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (!Params.AddRoutingConfig)  return String.Empty; #>
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web;
using System.Web.Mvc;
using System.Web.Routing;
using ITHit.WebDav.Server.SampleModel.Config;

namespace ITHit.WebDav.Server.SampleModel
{
    /// <summary>
    /// This module is needed for ASP.NET MVC application to forbid routing
    /// of requests that should go DavHandler.
    /// It reads DavLocations section and inserts ignore rules for urls that correspond
    /// to these locations.
    /// </summary>
    public class UrlRoutingModule : IHttpModule
    {
        void IHttpModule.Init(HttpApplication application)
        {
            var davLocationsSection = DavLocationsSection.Instance;
            var l = new List<RouteBase>(RouteTable.Routes);
            RouteTable.Routes.Clear();
            foreach (DavLocation location in davLocationsSection.Locations)
            {
                var methods = location.Verbs.Split(new[] { "," }, StringSplitOptions.RemoveEmptyEntries)
                    .Select(verb => verb.Trim().ToUpper()).ToArray();

                var prefix = location.Path.Trim(new[] { '/', '\\' });
                var pat = (prefix == string.Empty ? string.Empty : prefix + "/") + "{*rest}";
                if (methods.Length > 0)
                {
                    RouteTable.Routes.IgnoreRoute(pat, new { httpMethod = new HttpMethodConstraint(methods) });
                }
                else
                {
                    RouteTable.Routes.IgnoreRoute(pat);
                }
            }

            foreach (var route in l)
            {
                RouteTable.Routes.Add(route);
            }
        }

        public void Dispose()
        {
        }
    }
}
