//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Storage != StorageType.Ntfs)  return string.Empty; #>
using System;
using System.Collections.Generic;
//!<# if (Params.ACL) { #>
using System.DirectoryServices.AccountManagement;
//!<# } #>
using System.IO;
using System.Linq;
//!<# if (Params.ACL) { #>
using System.Security.AccessControl;
using System.Security.Principal;
//!<# } #>

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Acl;
//!<# if (Params.Class2) { #>
using ITHit.WebDAV.Server.Class2;
//!<# } #>
//!<# if (Params.MicrosoftExtensions) { #>
using ITHit.WebDAV.Server.MicrosoftExtensions;
//!<# } #>
//!<# if (Params.ACL) { #>
using ITHit.WebDAV.Server.Quota;
using ITHit.WebDav.Server.SampleModel.__NtfsStorage__.Acl;
//!<# } #>
using ITHit.WebDav.Server.SampleModel.__NtfsStorage__.NtfsStreamsEngine;

//!<# if (Params.ACL) { #>
using IPrincipal = ITHit.WebDAV.Server.Acl.IPrincipal;
//!<# } #>

namespace ITHit.WebDav.Server.SampleModel.__NtfsStorage__
{
    /// <summary>
    /// Base class for items(folders, files) in this sample.
    /// </summary>
    public abstract class DavHierarchyItem : IHierarchyItem/*!<# if (Params.ACL) { #>!*/, IAclHierarchyItem/*!<# } #>!*//*!<# if (Params.Class2) { #>!*/, ILock/*!<# } #>!*//*!<# if (Params.MicrosoftExtensions) { #>!*/, IMsItem/*!<# } #>!*/
    {
        /// <summary>
        /// Instance of <see cref="FileSystemInfo"/> corresponding to a file represented by this object.
        /// </summary>
        internal FileSystemInfo fileSystemInfo;

        /// <summary>
        /// Instance of <see cref="DavContext"/>.
        /// </summary>
        protected DavContext context;

        /// <summary>
        /// Encoded WebDAV relative path to the item.
        /// </summary>
        private readonly string path;

        /// <summary>
        /// User defined property values.
        /// </summary>
        private List<PropertyValue> propertyValues;

        /// <summary>
        /// Locks acquired on the item.
        /// </summary>
        private List<DateLockInfo> explicitLocks;

        /// <summary>
        /// Initializes a new instance of the DavHierarchyItem class.
        /// </summary>
        /// <param name="fileSystemInfo">Corresponding <see cref="FileSystemInfo"/> to this file/folder.</param>
        /// <param name="context">Instance of <see cref="DavContext"/>.</param>
        /// <param name="parentPath">Encoded relative WebDAV path to parent folder.</param>
        protected DavHierarchyItem(FileSystemInfo fileSystemInfo, DavContext context, string parentPath)
        {
            this.fileSystemInfo = fileSystemInfo;
            this.context = context;
            this.path = parentPath;
        }

        //$<IHierarchyItem.Name
        /// <summary>
        /// Gets name of the item.
        /// </summary>
        public string Name
        {
            get { return fileSystemInfo.Name; }
        }
        //$>

        //$<IHierarchyItem.Created
        /// <summary>
        /// Gets date when the item was created in UTC.
        /// </summary>
        public DateTime Created
        {
            get { return fileSystemInfo.CreationTimeUtc; }
        }
        //$>

        //$<IHierarchyItem.Modified
        /// <summary>
        /// Gets date when the item was last modified in UTC.
        /// </summary>
        public DateTime Modified
        {
            get { return fileSystemInfo.LastWriteTimeUtc; }
        }
        //$>

        //$<IHierarchyItem.Path
        /// <summary>
        /// Gets path of the item where each part between slashes is encoded.
        /// </summary>
        public virtual string Path
        {
            get { return path; }
        }
        //$>

        /// <summary>
        /// Returns parent folder of the item.
        /// </summary>
        /// <returns>Parent folder.</returns>
        public DavFolder GetParent()
        {
            //Calculate local file system path of parent item.
            string parentLocalPath = fileSystemInfo.FullName.TrimEnd('\\');
            int index = parentLocalPath.LastIndexOf('\\');
            if (index <= context.RepositoryPath.Length - 1)
            {
                return null;
            }

            var directory = new DirectoryInfo(parentLocalPath.Remove(index));

            //Calculate WebDAV root relative parent path.
            string parentPath = Path.TrimEnd('/');
            int ind = parentPath.LastIndexOf('/');
            if (ind >= 0)
            {
                parentPath = parentPath.Substring(0, ind);
            }

            return new DavFolder(directory, context, parentPath);
        }

        /// <summary>
        /// Creates a copy of this item with a new name in the destination folder.
        /// </summary>
        /// <param name="destFolder">Destination folder.</param>
        /// <param name="destName">Name of the destination item.</param>
        /// <param name="deep">Indicates whether to copy entire subtree.</param>
        /// <param name="multistatus">If some items fail to copy but operation in whole shall be continued, add
        /// information about the error into <paramref name="multistatus"/> using 
        /// <see cref="MultistatusException.AddInnerException(string,ITHit.WebDAV.Server.DavException)"/>.
        /// </param>
        public abstract void CopyTo(
            IItemCollection destFolder,
            string destName,
            bool deep,
            MultistatusException multistatus);

        /// <summary>
        /// Moves this item to the destination folder under a new name.
        /// </summary>
        /// <param name="destFolder">Destination folder.</param>
        /// <param name="destName">Name of the destination item.</param>
        /// <param name="multistatus">If some items fail to copy but operation in whole shall be continued, add
        /// information about the error into <paramref name="multistatus"/> using 
        /// <see cref="MultistatusException.AddInnerException(string,ITHit.WebDAV.Server.DavException)"/>.
        /// </param>
        public abstract void MoveTo(IItemCollection destFolder, string destName, MultistatusException multistatus);

        /// <summary>
        /// Deletes this item.
        /// </summary>
        /// <param name="multistatus">If some items fail to delete but operation in whole shall be continued, add
        /// information about the error into <paramref name="multistatus"/> using
        /// <see cref="MultistatusException.AddInnerException(string,ITHit.WebDAV.Server.DavException)"/>.
        /// </param>
        public abstract void Delete(MultistatusException multistatus);

        //$<IHierarchyItem.GetProperties
        /// <summary>
        /// Retrieves user defined property values.
        /// </summary>
        /// <param name="names">Names of dead properties which values to retrieve.</param>
        /// <param name="allprop">Whether all properties shall be retrieved.</param>
        /// <returns>Property values.</returns>
        public IEnumerable<PropertyValue> GetProperties(IList<PropertyName> names, bool allprop)
        {
            if (allprop)
            {
                // get all properties
                return getPropertyValues();
            }

            //Get requeste property values.
            return getPropertyValues().Where(property => names.Contains(property.QualifiedName));
        }
        //$>

        //$<IHierarchyItem.GetPropertyNames
        /// <summary>
        /// Retrieves names of all user defined properties.
        /// </summary>
        /// <returns>Property names.</returns>
        public IEnumerable<PropertyName> GetPropertyNames()
        {
            return getPropertyValues().Select(p => p.QualifiedName);
        }
        //$>

        //!<# if (Params.Class2) { #>
        /// <summary>
        /// Check that if the item is locked then client has submitted correct lock token.
        /// </summary>
        public void RequireHasToken()
        {
            var allLocks = getExplicitLocks().ToList();
            if (allLocks.Count == 0)
            {
                return;
            }

            IList<string> clientLockTokens = context.Request.ClientLockTokens;
            if (allLocks.Where(l => clientLockTokens.Contains(l.LockToken)).Any())
            {
                return;
            }

            throw new LockedException();
        }

        /// <summary>
        /// Ensure that there are no active locks on the item.
        /// </summary>
        /// <param name="skipShared">Whether shared locks shall be checked.</param>
        public void RequireUnlocked(bool skipShared)
        {
            if (skipShared && getExplicitLocks().Where(l => l.Level == LockLevel.Exclusive).Any())
            {
                throw new LockedException();
            }

            if (!skipShared && getExplicitLocks().Any())
            {
                throw new LockedException();
            }
        }
        //!<# } #>

        //$<IHierarchyItem.UpdateProperties
        /// <summary>
        /// Saves dead property values to NTFS alternate data stream.
        /// </summary>
        /// <param name="setProps">Properties to be set.</param>
        /// <param name="delProps">Properties to be deleted.</param>
        /// <param name="multistatus">We don't use it here.</param>
        public virtual void UpdateProperties(
            IList<PropertyValue> setProps,
            IList<PropertyName> delProps,
            MultistatusException multistatus)
        {
            //!<# if (Params.Class2) { #>
            RequireHasToken();

            //!<# } #>
            foreach (PropertyValue propToSet in setProps)
            {
                // Microsoft Mini-redirector may update file creation date, modification date and access time passing properties:
                // <Win32CreationTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:15:34 GMT</Win32CreationTime>
                // <Win32LastModifiedTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:36:24 GMT</Win32LastModifiedTime>
                // <Win32LastAccessTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:36:24 GMT</Win32LastAccessTime>
                // In this case update creation and modified date in your storage or do not save this properties at all, otherwise 
                // Windows Explorer will display creation and modification date from this props and it will differ from the values 
                // in the Created and Modified fields in your storage 
                if (propToSet.QualifiedName.Namespace == "urn:schemas-microsoft-com:")
                {
                    if (propToSet.QualifiedName.Name == "Win32CreationTime")
                    {
                        fileSystemInfo.CreationTimeUtc =
                            DateTime.Parse(propToSet.Value, new System.Globalization.CultureInfo("en-US")).ToUniversalTime();
                    }
                    else if (propToSet.QualifiedName.Name == "Win32LastModifiedTime")
                    {
                        fileSystemInfo.LastWriteTimeUtc =
                            DateTime.Parse(propToSet.Value, new System.Globalization.CultureInfo("en-US")).ToUniversalTime();
                    }
                }
                else
                {
                    var existingProp = getPropertyValues()
                        .Where(p => p.QualifiedName == propToSet.QualifiedName).FirstOrDefault();

                    if (existingProp != null)
                    {
                        existingProp.Value = propToSet.Value;
                    }
                    else
                    {
                        getPropertyValues().Add(propToSet);
                    }
                }
            }

            getPropertyValues().RemoveAll(prop => delProps.Contains(prop.QualifiedName));

            context.FileOperation(
                this,
                () => RewriteStream("Properties", propertyValues),
                Privilege.Write);

            // You should not update modification date/time here. Mac OS X Finder expects that properties update do not change the file modification date.
        }
        //$>
        
        //!<# if (Params.Class2) { #>
        //$<ILock.GetActiveLocks
        /// <summary>
        /// Retrieves non expired locks for this item.
        /// </summary>
        /// <returns>Enumerable with information about locks.</returns>
        public IEnumerable<LockInfo> GetActiveLocks()
        {
            //!<# if (Params.ACL) { #>
            RequireReadPrivilege();
            //!<# } #>
            return getExplicitLocks().Select(
                    l => new LockInfo
                         {
                             IsDeep = l.IsDeep,
                             Level = l.Level,
                             Owner = l.ClientOwner,
                             LockRoot = l.LockRoot,
                             TimeOut = l.Expiration == DateTime.MaxValue ?
                                TimeSpan.MaxValue :
                                l.Expiration - DateTime.UtcNow,
                             Token = l.LockToken
                         });
        }
        //$>

        //$<ILock.Lock

        /// <summary>
        /// Locks this item.
        /// </summary>
        /// <param name="level">Whether lock is share or exclusive</param>
        /// <param name="isDeep">Whether lock is deep.</param>
        /// <param name="requestedTimeOut">Lock timeout which was requested by client.
        /// Server may ignore this parameter and set any timeout.</param>
        /// <param name="owner">Owner of the lock as specified by client.</param> 
        /// <returns>
        /// Instance of <see cref="LockResult"/> with information about the lock.
        /// </returns>
        public LockResult Lock(LockLevel level, bool isDeep, TimeSpan? requestedTimeOut, string owner)
        {
            RequireUnlocked(level == LockLevel.Shared);

            string token = Guid.NewGuid().ToString();

            TimeSpan timeOut;
            if (!requestedTimeOut.HasValue || requestedTimeOut == TimeSpan.MaxValue)
            {
                // If timeout is absent or infinit timeout requested,
                // grant 5 minute lock.
                timeOut = TimeSpan.FromMinutes(5);
            }
            else
            {
                timeOut = requestedTimeOut.Value;
            }

            DateTime expiration = DateTime.UtcNow + timeOut;

            explicitLocks.Add(new DateLockInfo
                                  {
                                      Expiration = expiration,
                                      IsDeep = false,
                                      Level = level,
                                      LockRoot = Path,
                                      LockToken = token,
                                      ClientOwner = owner,
                                      //!<# if (Params.ACL) { #>
                                      OwnerUser = context.User.Name,
                                      //!<# } #>
                                      TimeOut = timeOut
                                  });
            saveLocks();

            return new LockResult(token, timeOut);
        }
        //$>

        //$<ILock.RefreshLock
        /// <summary>
        /// Updates lock timeout information on this item.
        /// </summary>
        /// <param name="token">Lock token.</param>
        /// <param name="requestedTimeOut">Lock timeout which was requested by client.
        /// Server may ignore this parameter and set any timeout.</param>
        /// <returns>
        /// Instance of <see cref="LockResult"/> with information about the lock.
        /// </returns>
        public RefreshLockResult RefreshLock(string token, TimeSpan? requestedTimeOut)
        {
            DateLockInfo li = getExplicitLocks().Where(l => l.LockToken == token).SingleOrDefault();

            if (li == null)
            {
                throw new DavException("Lock can not be found.", DavStatus.PRECONDITION_FAILED);
            }

            if (requestedTimeOut.HasValue && requestedTimeOut != TimeSpan.MaxValue)
            {
                // Update timeout if it is specified and not Infinity.
                // Otherwise leave previous timeout.
                li.TimeOut = requestedTimeOut.Value;
            }

            if (!requestedTimeOut.HasValue)
                li.TimeOut = TimeSpan.FromMinutes(5);

            li.Expiration = DateTime.UtcNow + li.TimeOut;

            saveLocks();

            return new RefreshLockResult(li.Level, li.IsDeep, li.TimeOut, li.ClientOwner);
        }
        //$>

        //$<ILock.Unlock
        /// <summary>
        /// Removes lock with the specified token from this item.
        /// </summary>
        /// <param name="lockToken">Lock with this token should be removed from the item.</param>
        public void Unlock(string lockToken)
        {
            int i = getExplicitLocks().FindIndex(li => li.LockToken == lockToken);
            if (i >= 0)
            {
                //!<#if (Params.ACL) { #>
                var acl = File.GetAccessControl(
                    fileSystemInfo.FullName,
                    AccessControlSections.Access | AccessControlSections.Owner | AccessControlSections.Group);

                if (getExplicitLocks()[i].OwnerUser == context.User.Name || 
                    EffectivePermissions.GetEffectivePermissions(context.User, acl) == FileSystemRights.FullControl)
                {
                    getExplicitLocks().RemoveAt(i);
                }
                else
                {
                    throw new NeedPrivilegesException(
                        "Only owner or Administrators can unlock resources",
                        Path,
                        Privilege.Unlock);
                }
                //!<# } #>

                //!<#if (!Params.ACL) { #>
                getExplicitLocks().RemoveAt(i);
                //!<# } #>
            }
            else
            {
                throw new DavException("The lock could not be found.", DavStatus.PRECONDITION_FAILED);
            }

            saveLocks();
        }
        //$>
        //!<# } #>

        //!<# if (Params.ACL) { #>
        /// <summary>
        /// Creates <see cref="Group"/> or <see cref="User"/> for windows SID.
        /// </summary>
        /// <param name="sid">Windows SID.</param>
        /// <param name="context">Instance of <see cref="DavContext"/>.</param>
        /// <returns>Corresponding <see cref="User"/> or <see cref="Group"/> or <c>null</c> if there's no user
        /// or group which correspond to specified sid.</returns>
        public static IPrincipal FromSid(string sid, DavContext context)
        {
            Principal pr = Principal.FindByIdentity(context.GetPrincipalContext(), IdentityType.Sid, sid);
            if (pr == null)
            {
                return null;
            }

            return pr is GroupPrincipal
                       ? (IPrincipal)new Group((GroupPrincipal)pr, context)
                       : new User((UserPrincipal)pr, context);
        }

        //$<IAclHierarchyItem.SetOwner
        /// <summary>
        /// Sets <paramref name="principal"/> as new owner for this item.
        /// </summary>
        /// <param name="principal">New owner.</param>
        public void SetOwner(IPrincipal principal)
        {
            context.FileOperation(
                this,
                () =>
                    {
                        var accessControl = File.GetAccessControl(fileSystemInfo.FullName, AccessControlSections.Owner);
                        accessControl.SetOwner(((PrincipalBase)principal).Sid);
                        File.SetAccessControl(fileSystemInfo.FullName, accessControl);
                    },
                Privilege.WriteProperties);
        }
        //$>

        //$<IAclHierarchyItem.GetOwner
        /// <summary>
        /// Retrieves user or group who is owner for this item.
        /// </summary>
        /// <returns>Owner for this item.</returns>
        public IPrincipal GetOwner()
        {
            return context.FileOperation(
                this,
                () =>
                    {
                        FileSecurity acl = File.GetAccessControl(fileSystemInfo.FullName);
                        return FromSid(acl.GetOwner(typeof(SecurityIdentifier)).Value, context);
                    },
                Privilege.Read);
        }
        //$>

        //$<IAclHierarchyItem.SetGroup
        /// <summary>
        /// File groups are not supported in Windows.
        /// </summary>
        public void SetGroup(IPrincipal principal)
        {
            throw new DavException("Group cannot be set", DavStatus.FORBIDDEN);
        }
        //$>

        //$<IAclHierarchyItem.GetGroup
        /// <summary>
        /// File groups are not supported in Windows.
        /// </summary>
        public IPrincipal GetGroup()
        {
            return null;
        }
        //$>

        //$<IAclHierarchyItem.GetSupportedPrivilegeSet
        /// <summary>
        /// Retrieves list of all privileges (permissions) which can be set for the item.
        /// </summary>
        /// <returns>Enumerable with supported permissions.</returns>
        public IEnumerable<SupportedPrivilege> GetSupportedPrivilegeSet()
        {
            RequireReadPrivilege();
            return NtfsPermissions.GetPermissionsTree();
        }
        //$>

        //$<IAclHierarchyItem.GetCurrentUserPrivilegeSet
        /// <summary>
        /// Retrieves the exact set of permissions current user has for this file or folder.
        /// </summary>
        public IEnumerable<Privilege> GetCurrentUserPrivilegeSet()
        {
            var acl = File.GetAccessControl(
                fileSystemInfo.FullName,
                AccessControlSections.Access | AccessControlSections.Owner | AccessControlSections.Group);

            return context.FileOperation(
                this,
                () => NtfsPermissions.ExpandPrivileges(
                    NtfsPermissions.MapFileSystemRights(
                        EffectivePermissions.GetEffectivePermissions(context.User, acl))),
                Privilege.ReadAcl);
        }
        //$>

        //$<IAclHierarchyItem.GetAcl
        /// <summary>
        /// Retrieves access control list for this file or folder.
        /// </summary>
        /// <param name="propertyNames">Properties which will be retrieved from users/groups specified in
        /// access control list returned.</param>
        /// <returns>Enumerable with access control entries.</returns>
        public IEnumerable<ReadAce> GetAcl(IList<PropertyName> propertyNames)
        {
            AuthorizationRuleCollection rules = null;
            AuthorizationRuleCollection protectedRules = null;
            context.FileOperation(
                this,
                () =>
                    {
                        FileSecurity acl = File.GetAccessControl(fileSystemInfo.FullName);
                        rules = acl.GetAccessRules(true, false, typeof(SecurityIdentifier));
                        protectedRules = acl.GetAccessRules(false, true, typeof(SecurityIdentifier));
                    },
                Privilege.ReadAcl);

            foreach (FileSystemAccessRule rule in rules)
            {
                var ace = new ReadAce();
                ace.Principal = FromSid(rule.IdentityReference.Value, context);
                if (ace.Principal == null)
                {
                    continue;
                }

                if (rule.AccessControlType == AccessControlType.Allow)
                {
                    ace.GrantPrivileges.AddRange(NtfsPermissions.MapFileSystemRights(rule.FileSystemRights));
                }
                else
                {
                    ace.DenyPrivileges.AddRange(NtfsPermissions.MapFileSystemRights(rule.FileSystemRights));
                }

                ace.IsProtected = false;
                yield return ace;
            }

            foreach (FileSystemAccessRule rule in protectedRules)
            {
                var ace = new ReadAce();
                ace.Principal = FromSid(rule.IdentityReference.Value, context);
                if (ace.Principal == null)
                {
                    context.Logger.LogDebug("Could not resolve principal with sid " + rule.IdentityReference.Value);
                    continue;
                }

                if (rule.AccessControlType == AccessControlType.Allow)
                {
                    ace.GrantPrivileges.AddRange(NtfsPermissions.MapFileSystemRights(rule.FileSystemRights));
                }
                else
                {
                    ace.DenyPrivileges.AddRange(NtfsPermissions.MapFileSystemRights(rule.FileSystemRights));
                }

                ace.IsProtected = true;
                ace.InheritedFrom = GetParent() ?? this;
                yield return ace;
            }
        }
        //$>

        //$<IAclHierarchyItem.SetAcl
        /// <summary>
        /// Sets access control list for this file/folder.
        /// </summary>
        /// <param name="aces">List of access control entries.</param>
        public void SetAcl(IList<WriteAce> aces)
        {
            FileSecurity acl = File.GetAccessControl(fileSystemInfo.FullName);
            var rules = acl.GetAccessRules(true, false, typeof(SecurityIdentifier));
            foreach (FileSystemAccessRule rule in rules)
            {
                //remove only those records we can resolve (in our domain/computer)
                if (FromSid(rule.IdentityReference.Value, context) != null)
                {
                   acl.RemoveAccessRule(rule);
                }
            }

            foreach (WriteAce ace in aces)
            {
                FileSystemRights grantRights = NtfsPermissions.MapPrivileges(ace.GrantPrivileges, context.Logger);
                var sid = ((PrincipalBase)ace.Principal).Sid;
                if (grantRights != 0)
                {
                    var grantRule = new FileSystemAccessRule(sid, grantRights, AccessControlType.Allow);
                    acl.AddAccessRule(grantRule);
                }

                FileSystemRights denyRights = NtfsPermissions.MapPrivileges(ace.DenyPrivileges, context.Logger);
                if (denyRights != 0)
                {
                    var denyRule = new FileSystemAccessRule(sid, denyRights, AccessControlType.Deny);
                    acl.AddAccessRule(denyRule);
                }
            }

            context.FileOperation(
                this,
                () => File.SetAccessControl(fileSystemInfo.FullName, acl),
                Privilege.WriteProperties);
        }
        //$>

        //$<IAclHierarchyItem.GetAclRestrictions
        /// <summary>
        /// Retrieves list of restrictions for access control entries.
        /// We don't support WebDAV inverted permissions.
        /// </summary>
        /// <returns>ACL restrictions.</returns>
        public AclRestriction GetAclRestrictions()
        {
            return new AclRestriction { NoInvert = true };
        }
        //$>

        //$<IAclHierarchyItem.GetInheritedAclSet
        /// <summary>
        /// Gets all folders, from which this file/folder has inherited access control entries.
        /// </summary>
        /// <returns>Enumerable with files/folders from which this file/folder has inherited
        /// access control entries.</returns>
        public IEnumerable<IHierarchyItem> GetInheritedAclSet()
        {
            var current = this;
            var parent = current.GetParent();
            while (parent != null)
            {
                var accessControl = File.GetAccessControl(parent.fileSystemInfo.FullName);
                var acl = accessControl.GetAccessRules(true, false, typeof(SecurityIdentifier));
                if (acl.Count > 0)
                {
                    yield return parent;
                }

                current = parent;
                parent = current.GetParent();
            }
        }
        //$>

        //$<IAclHierarchyItem.GetPrincipalCollectionSet
        /// <summary>
        /// Gets collections which contain principals.
        /// We return one folder which contains users and the other folder which contains groups.
        /// We put them into different folders because webdav doesn't have strong distinction between users
        /// and groups, but Windows does, so we put users and groups into separate virtual folders so it is obvious
        /// what user wants to do (create user or group for example).
        /// </summary>
        /// <returns>Folders which contain users/groups.</returns>
        public IEnumerable<IPrincipalFolder> GetPrincipalCollectionSet()
        {
            return new IPrincipalFolder[]
                       {
                           new UserFolder(context),
                           new GroupFolder(context)
                       };
        }
        //$>

        //$<IAclHierarchyItem.ResolveWellKnownPrincipal
        /// <summary>
        /// Retrieves user or group which correspond to a well known principal
        /// (defined in <see cref="WellKnownPrincipal"/>.)
        /// </summary>
        /// <param name="wellKnownPrincipal">Well known principal type.</param>
        /// <returns>Instance of corresponding user/group or <c>null</c> if corresponding user/group
        /// is not supported.</returns>
        public IPrincipal ResolveWellKnownPrincipal(WellKnownPrincipal wellKnownPrincipal)
        {
            switch (wellKnownPrincipal)
            {
                case WellKnownPrincipal.All:
                    var sid1 = new SecurityIdentifier(WellKnownSidType.WorldSid, null);
                    return FromSid(sid1.Value, context);
                case WellKnownPrincipal.Authenticated:
                    var sid2 = new SecurityIdentifier(WellKnownSidType.AuthenticatedUserSid, null);
                    return FromSid(sid2.Value, context);
                case WellKnownPrincipal.Unauthenticated:
                    return context.AnonymousUser != null ? FromSid(context.AnonymousUser.User.Value, context) : null;
                case WellKnownPrincipal.Self:
                    return null;
            }

            return null;
        }
        //$>

        //$<IAclHierarchyItem.GetItemsByProperty
        /// <summary>
        /// Find all resources in the subtree which have either Group or Owner (defined by <paramref name="matchBy"/>
        /// parameter) which identifies current user or group current user belongs to.
        /// For example, this report can return all of the resources in a collection
        /// hierarchy that are owned by the current user.
        /// We don't implement it here as all implementatino is in <see cref="DavFolder.GetItemsByProperty"/>.
        /// </summary>
        public virtual IEnumerable<IAclHierarchyItem> GetItemsByProperty(MatchBy matchBy, IList<PropertyName> props)
        {
            yield break;
        }
        //$>

        //$<IAclHierarchyItem.GetCurrentUserPrincipal
        /// <summary>
        /// Returns instance of <see cref="IPrincipal"/> which represents current user.
        /// </summary>
        /// <returns>Current user.</returns>
        public IPrincipal GetCurrentUserPrincipal()
        {
            return FromSid(context.User.User.Value, context);
        }
        //$>

        /// <summary>
        /// Throws <see cref="NeedPrivilegesException"/> if current user doesn't have Read permission on this
        /// file/folder.
        /// </summary>
        protected void RequireReadPrivilege()
        {
            var acl = File.GetAccessControl(
                fileSystemInfo.FullName,
                AccessControlSections.Access | AccessControlSections.Owner | AccessControlSections.Group);

            var rights = EffectivePermissions.GetEffectivePermissions(context.User, acl);
            if ((rights & FileSystemRights.Read) != FileSystemRights.Read)
            {
                throw new NeedPrivilegesException("Missing read privilege", Path, Privilege.Read);
            }
        }
        //!<# } #>

        //!<# if (Params.MicrosoftExtensions) { #>
        //$<IMsItem.GetFileAttributes
        /// <summary>
        /// Returns Windows file attributes (readonly, hidden etc.) for this file/folder.
        /// </summary>
        /// <returns>Windows file attributes.</returns>
        public FileAttributes GetFileAttributes()
        {
            //!<# if (Params.MarkMacOsFilesHidden) { #>
            if (Name.StartsWith("."))
            {
                return fileSystemInfo.Attributes | FileAttributes.Hidden;
            }
            //!<# } #>
            return fileSystemInfo.Attributes;
        }
        //$>

        //$<IMsItem.SetFileAttributes
        /// <summary>
        /// Sets Windows file attributes (readonly, hidden etc.) on this item.
        /// </summary>
        /// <param name="value">File attributes.</param>
        public void SetFileAttributes(FileAttributes value)
        {
            File.SetAttributes(fileSystemInfo.FullName, value);
        }
        //$>
        //!<# } #>

        /// <summary>
        /// Retrieves full local path for this file/folder.
        /// </summary>
        /// <returns>Full local path.</returns>
        internal string GetFullPath()
        {
            return fileSystemInfo.FullName.TrimEnd('\\');
        }

        /// <summary>
        /// Returns path of the item in the hierarchy tree.
        /// </summary>
        internal string GetHierarchyPath()
        {
            int index = context.RepositoryPath.Length;
            if (GetFullPath().Length > index)
            {
                index++;
            }

            return GetFullPath().Remove(0, index).Replace('\\', '/');
        }

        /// <summary>
        /// Gets NTFS stream with specified name and deserializes object of type <typeparamref name="T"/>
        /// from it.
        /// </summary>
        /// <typeparam name="T">Type of object serialized in stream.</typeparam>
        /// <param name="streamName">Name of NTFS stream stored as NTFS stream in this file/folder.</param>
        /// <param name="logger">Instance of <see cref="logger"/>.</param>
        /// <returns>Deserialized object.</returns>
        internal T GetStreamAndDeserialize<T>(string streamName, ILogger logger) where T : new()
        {
            return context.FileOperation(
                this,
                () => SerializationUtils.GetStreamAndDeserialize<T>(streamName, fileSystemInfo, logger),
                Privilege.Read);
        }

        /// <summary>
        /// Serializes object <paramref name="o"/> and saves it into specified NTFS stream.
        /// </summary>
        /// <typeparam name="T">Type of object to be serialized.</typeparam>
        /// <param name="streamName">NTFS stream name.</param>
        /// <param name="o">Object to be serialized.</param>
        protected abstract void RewriteStream<T>(string streamName, T o);

        /// <summary>
        /// Deletes NTFS stream with specified name.
        /// </summary>
        /// <param name="fileName">File name.</param>
        /// <param name="streamName">NTFS stream name.</param>
        protected abstract void DeleteStream(string fileName, string streamName);

        /// <summary>
        /// Creates <see cref="DavFolder"/> with local path <see cref="localPath"/> and encoded WebDAV relative path
        /// <see cref="path"/>.
        /// </summary>
        /// <param name="localPath">Full local folder path.</param>
        /// <param name="webDavPath">Encoded WebDAV relative path for this folder.</param>
        /// <returns>Instance of <see cref="DavFolder"/>.</returns>
        protected DavFolder GetDavFolder(string localPath, string webDavPath)
        {
            return new DavFolder(new DirectoryInfo(localPath), context, webDavPath);
        }

        /// <summary>
        /// Creates <see cref="DavFile"/> with local path <see cref="localPath"/> and encoded WebDAV relative path
        /// <see cref="path"/>.
        /// </summary>
        /// <param name="localPath">Full local folder path.</param>
        /// <param name="webDavPath">Encoded WebDAV relative path for this folder.</param>
        /// <returns>Instance of <see cref="DavFolder"/>.</returns>
        protected DavFile GetDavFile(string localPath, string webDavPath)
        {
            return new DavFile(new FileInfo(localPath), context, webDavPath);
        }

        /// <summary>
        /// Retrieves non-expired locks acquired on this item.
        /// </summary>
        /// <returns>List of locks with their expiration dates.</returns>
        private List<DateLockInfo> getExplicitLocks()
        {
            if (explicitLocks == null)
            {
                explicitLocks = GetStreamAndDeserialize<List<DateLockInfo>>("Locks", context.Logger);
                explicitLocks.RemoveAll(li => li.Expiration <= DateTime.UtcNow);
                explicitLocks.ForEach(l => l.LockRoot = Path);
            }

            return explicitLocks;
        }

        /// <summary>
        /// Retrieves list of user defined propeties for this item.
        /// </summary>
        /// <returns>List of user defined properties.</returns>
        private List<PropertyValue> getPropertyValues()
        {
            if (propertyValues == null)
            {
                // Deserialize property values from alternate data stream.
                propertyValues = GetStreamAndDeserialize<List<PropertyValue>>("Properties", context.Logger);
            }

            return propertyValues;
        }

        /// <summary>
        /// Saves locks acquired on this file/folder in NTFS stream.
        /// </summary>
        private void saveLocks()
        {
            if (getExplicitLocks().Count > 0)
            {
                RewriteStream("Locks", explicitLocks);
            }
            else
            {
                DeleteStream(fileSystemInfo.FullName, "Locks");
            }
        }
    }
}
