//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Storage != StorageType.Db)  return string.Empty; #>
using System;
using System.Collections;
using System.Collections.Generic;
//!<# if (Params.MicrosoftExtensions) { #>
using System.IO;
//!<# } #>
using System.Linq;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Class1;
//!<# if (Params.Class2) { #>
using ITHit.WebDAV.Server.Class2;
//!<# } #>
//!<# if (Params.MicrosoftExtensions) { #>
using ITHit.WebDAV.Server.MicrosoftExtensions;
//!<# } #>

namespace ITHit.WebDav.Server.SampleModel.__SqlStorage__
{
    /// <summary>
    /// Base class for items like files, folders, versions etc.
    /// </summary>
    public abstract class DavHierarchyItem : IHierarchyItem/*!<# if (Params.Class2) { #>!*/, ILock/*!<# } #>!*//*!<# if (Params.MicrosoftExtensions) { #>!*/, IMsItem/*!<# } #>!*/
    {
        protected DavContext Context { get; private set; }

        public Guid ItemId { get; private set; }

        public string Name { get; private set; }

        public string Path { get; private set; }

        public DateTime Created { get; private set; }

        public DateTime Modified { get; private set; }

        protected Guid ParentId { get; private set; }

        //!<# if (Params.MicrosoftExtensions) { #>
        private FileAttributes fileAttributes;
        //!<# } #>

        public DavHierarchyItem(
            DavContext context,
            Guid itemId,
            Guid parentId,
            string name,
            string path,
            DateTime created,
            DateTime modified/*!<# if (Params.MicrosoftExtensions) { #>!*/, /*!<#}#>!*/
            /*!<# if (Params.MicrosoftExtensions) { #>!*/FileAttributes fileAttributes/*!<#}#>!*/)
        {
            Context = context;
            ItemId = itemId;
            ParentId = parentId;
            Name = name;
            Path = path;
            Created = created;
            Modified = modified;
            //!<# if (Params.MicrosoftExtensions) { #>
            this.fileAttributes = fileAttributes;
            //!<# } #>
        }

        public DavFolder GetParent()
        {
            string[] parts = Path.TrimEnd('/').Split('/');
            string parentParentPath = "/";
            if (parts.Length > 2)
            {
                parentParentPath = string.Join("/", parts, 0, parts.Length - 2);
            }

            string command =
                @"SELECT 
                     ItemID
                   , ParentItemId
                   , ItemType
                   , Name
                   , Created
                   , Modified
                   <# if (Params.MicrosoftExtensions) { #>                   , FileAttributes<#}#>
                  FROM Item
                  WHERE ItemId = @ItemId";

            return Context.ExecuteItem<DavFolder>(
                parentParentPath,
                command,
                "@ItemId", ParentId).FirstOrDefault();
        }

        public abstract void CopyTo(
            IItemCollection destFolder,
            string destName,
            bool deep,
            MultistatusException multistatus);

        public abstract void MoveTo(IItemCollection destFolder, string destName, MultistatusException multistatus);

        public abstract void Delete(MultistatusException multistatus);

        public IEnumerable<PropertyValue> GetProperties(IList<PropertyName> names, bool allprop)
        {
            IList<PropertyValue> propVals =
                Context.ExecutePropertyValue(
                    "SELECT Name, Namespace, PropVal FROM Property WHERE ItemID = @ItemID",
                    "@ItemID", ItemId);

            foreach (PropertyValue p in propVals)
            {
                if (allprop || names.Contains(p.QualifiedName))
                {
                    yield return p;
                }
            }
        }
        
        public virtual void UpdateProperties(
            IList<PropertyValue> setProps,
            IList<PropertyName> delProps,
            MultistatusException multistatus)
        {
            //!<# if (Params.Class2) { #>
            RequireHasToken();
            //!<# } #>

            foreach (PropertyValue p in setProps)
            {
                // Microsoft Mini-redirector may update file creation date, modification date and access time passing properties:
                // <Win32CreationTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:15:34 GMT</Win32CreationTime>
                // <Win32LastModifiedTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:36:24 GMT</Win32LastModifiedTime>
                // <Win32LastAccessTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:36:24 GMT</Win32LastAccessTime>
                // In this case update creation and modified date in your storage or do not save this properties at all, otherwise 
                // Windows Explorer will display creation and modification date from this props and it will differ from the values 
                // in the Created and Modified fields in your storage 
                if (p.QualifiedName.Namespace == "urn:schemas-microsoft-com:")
                {
                    if (p.QualifiedName.Name == "Win32CreationTime")
                    {
                        SetDbField("Created", DateTime.Parse(p.Value, new System.Globalization.CultureInfo("en-US")).ToUniversalTime());
                    }
                    else if (p.QualifiedName.Name == "Win32LastModifiedTime")
                    {
                        SetDbField("Modified", DateTime.Parse(p.Value, new System.Globalization.CultureInfo("en-US")).ToUniversalTime());
                    }
                }
                else
                {
                    SetProperty(p); // create or update property
                }
            }

            foreach (PropertyName p in delProps)
            {
               RemoveProperty(p.Name, p.Namespace);
            }

            // You should not update modification date/time here. Mac OS X Finder expects that properties update do not change the file modification date.
        }

        public IEnumerable<PropertyName> GetPropertyNames()
        {
            foreach (PropertyValue val in GetProperties(new PropertyName[0], true))
            {
                yield return val.QualifiedName;
            }            
        }

        //!<# if (Params.Class2) { #>
        protected void RequireHasToken()
        {
            if (!ClientHasToken)
            {
                throw new LockedException();
            }
        }

        public LockResult Lock(LockLevel level, bool isDeep, TimeSpan? timeout, string owner)
        {
            if (ItemHasLock(level == LockLevel.Shared))
            {
                throw new LockedException();
            }

            if (isDeep)
            {
                // check if no items are locked in this subtree
                FindLocksDown(this, level == LockLevel.Shared);
            }

            if (!timeout.HasValue || timeout == TimeSpan.MaxValue)
            {
                // If timeout is absent or infinity timeout requested,
                // grant 5 minute lock.
                timeout = TimeSpan.FromMinutes(5);
            }

            // We store expiration time in UTC. If server/database is moved 
            // to other time zone the locks expiration time is always correct.
            DateTime expires = DateTime.UtcNow + timeout.Value;

            string token = Guid.NewGuid().ToString();
            string insertLockCommand = 
                @"INSERT INTO Lock (ItemID,Token,Shared,Deep,Expires,Owner)
                   VALUES(@ItemID, @Token, @Shared, @Deep, @Expires, @Owner)";

            Context.ExecuteNonQuery(
                insertLockCommand,
                "@ItemID", ItemId,
                "@Token", token,
                "@Shared", level == LockLevel.Shared,
                "@Deep", isDeep,
                "@Expires", expires,
                "@Owner", owner);

            return new LockResult(token, timeout.Value);
        }

        public RefreshLockResult RefreshLock(string token, TimeSpan? timeout)
        {
            LockInfo l = GetActiveLocks().Where(al => al.Token == token).FirstOrDefault();

            if (l == null)
            {
                throw new DavException("The lock doesn't exist", DavStatus.PRECONDITION_FAILED);
            }

            if (!timeout.HasValue || timeout == TimeSpan.MaxValue)
            {
                // If timeout is absent or infinity timeout requested,
                // grant 5 minute lock.
                l.TimeOut = TimeSpan.FromMinutes(5);
            }
            else
            {
                // Otherwise use new timeout.
                l.TimeOut = timeout.Value;
            }

            DateTime expires = DateTime.UtcNow + (TimeSpan)l.TimeOut;

            Context.ExecuteNonQuery(
                "UPDATE Lock SET Expires = @Expires WHERE Token = @Token",
                "@Expires", expires,
                "@Token", token);

            return new RefreshLockResult(l.Level, l.IsDeep, (TimeSpan)l.TimeOut, l.Owner);
        }

        public void Unlock(string lockToken)
        {
            if (!GetActiveLocks().Where(al => al.Token == lockToken).Any())
            {
                throw new DavException("This lock token doesn't correspond to any lock", DavStatus.PRECONDITION_FAILED);
            }

            // remove lock from existing item
            Context.ExecuteNonQuery(
                "DELETE FROM Lock WHERE Token = @Token",
                "@Token", lockToken);
        }

        public IEnumerable<LockInfo> GetActiveLocks()
        {
            Guid itemID = ItemId;
            List<LockInfo> l = new List<LockInfo>();

            l.AddRange(GetLocks(itemID, false)); // get all locks
            while (true)
            {
                itemID = Context.ExecuteScalar<Guid>(
                    "SELECT ParentItemId FROM Item WHERE ItemId = @ItemId",
                    "@ItemId", itemID);

                if (itemID == Guid.Empty)
                {
                    break;
                }

                l.AddRange(GetLocks(itemID, true)); // get only deep locks
            }

            return l;
        }
        //!<# } #>
        
        protected void SetProperty(PropertyValue prop)
        {
            string selectCommand = 
                @"SELECT Count(*) FROM Property
                  WHERE ItemID = @ItemID AND Name = @Name AND Namespace = @Namespace";

            int count = Context.ExecuteScalar<int>(
                selectCommand,
                "@ItemID", ItemId,
                "@Name", prop.QualifiedName.Name,
                "@Namespace", prop.QualifiedName.Namespace);

            // insert
            if (count == 0)
            {
                string insertCommand = @"INSERT INTO Property(ItemID, Name, Namespace, PropVal)
                                          VALUES(@ItemID, @Name, @Namespace, @PropVal)";

                Context.ExecuteNonQuery(
                    insertCommand,
                    "@PropVal", prop.Value,
                    "@ItemID", ItemId,
                    "@Name", prop.QualifiedName.Name,
                    "@Namespace", prop.QualifiedName.Namespace);
            }
            else
            {
                // update
                string command = @"UPDATE Property
                      SET PropVal = @PropVal
                      WHERE ItemID = @ItemID AND Name = @Name AND Namespace = @Namespace";

                Context.ExecuteNonQuery(
                    command,
                    "@PropVal", prop.Value,
                    "@ItemID", ItemId,
                    "@Name", prop.QualifiedName.Name,
                    "@Namespace", prop.QualifiedName.Namespace);
            }
        }

        protected void RemoveProperty(string name, string ns)
        {
            string command = @"DELETE FROM Property
                              WHERE ItemID = @ItemID
                              AND Name = @Name
                              AND Namespace = @Namespace";

            Context.ExecuteNonQuery(
                command,
                "@ItemID", ItemId,
                "@Name", name,
                "@Namespace", ns);
        }

        internal DavFolder CopyThisItem(DavFolder destFolder, DavHierarchyItem destItem, string destName)
        {
            // returns created folder, if any, otherwise null
            DavFolder createdFolder = null;

            Guid destID;
            if (destItem == null)
            {
                // copy item
                string commandText =
                    @"INSERT INTO Item(
                           ItemId
                         , Name
                         , Created
                         , Modified
                         , ParentItemId
                         , ItemType
                         , Content
                         , ContentType
                         , SerialNumber
                         //!<# if (Params.DeltaV) { #>
                         , CreatorDisplayName, Comment, AutoVersion, AutoCheckedOut
                         //!<# } #>
                         //!<# if (Params.ResumableUpload || Params.IEUpload) { #>
                         , TotalContentLength
                         , LastChunkSaved
                         //!<# } #>
                         //!<# if (Params.MicrosoftExtensions) { #>
                         , FileAttributes
                         //!<# } #>
                         )
                      SELECT
                           @Identity
                         , @Name
                         , GETUTCDATE()
                         , GETUTCDATE()
                         , @Parent
                         , ItemType
                         , Content
                         , ContentType
                         , SerialNumber
                         //!<# if (Params.DeltaV) { #>
                         , CreatorDisplayName
                         , Comment
                         , AutoVersion
                         , AutoCheckedOut
                         //!<# } #>
                         //!<# if (Params.ResumableUpload || Params.IEUpload) { #>
                         , TotalContentLength
                         , LastChunkSaved
                         //!<# } #>
                         //!<# if (Params.MicrosoftExtensions) { #>
                         , FileAttributes
                         //!<# } #>
                      FROM Item
                      WHERE ItemId = @ItemId";

                destID = Guid.NewGuid();
                Context.ExecuteNonQuery(
                    commandText,
                    "@Name", destName,
                    "@Parent", destFolder.ItemId,
                    "@ItemId", ItemId,
                    "@Identity", destID);

                destFolder.UpdateModified();

                if (this is IFolder)
                {
                    createdFolder = new DavFolder(
                        Context,
                        destID,
                        destFolder.ItemId,
                        destName,
                        destFolder.Path + EncodeUtil.EncodeUrlPart(destName) + "/",
                        DateTime.UtcNow,
                        DateTime.UtcNow/*!<# if (Params.MicrosoftExtensions) { #>!*/, /*!<#}#>!*/
                        /*!<# if (Params.MicrosoftExtensions) { #>!*/fileAttributes/*!<#}#>!*/);
                }
            }
            else 
            {
                // update existing destination
                destID = destItem.ItemId;

                string commandText = @"UPDATE Item SET
                                       Modified = GETUTCDATE()
                                       , ItemType = src.ItemType
                                       , ContentType = src.ContentType
                                       FROM (SELECT * FROM Item WHERE ItemId=@SrcID) src
                                       WHERE Item.ItemId=@DestID";

                Context.ExecuteNonQuery(
                    commandText,
                    "@SrcID", ItemId,
                    "@DestID", destID);

                // remove old properties from the destination
                Context.ExecuteNonQuery(
                    "DELETE FROM Property WHERE ItemID = @ItemID",
                    "@ItemID", destID);
            }

            // copy properties
            string command =
                @"INSERT INTO Property(ItemID, Name, Namespace, PropVal)
                  SELECT @DestID, Name, Namespace, PropVal
                  FROM Property
                  WHERE ItemID = @SrcID";

            Context.ExecuteNonQuery(
                command,
                "@SrcID", ItemId,
                "@DestID", destID);

            return createdFolder;
        }

        internal void MoveThisItem(DavFolder destFolder, string destName, DavFolder parent)
        {
            string command = 
                @"UPDATE Item SET
                     Name = @Name,
                     ParentItemId = @Parent
                  WHERE ItemId = @ItemID";

            Context.ExecuteNonQuery(
                command,
                "@ItemID", ItemId,
                "@Name", destName,
                "@Parent", destFolder.ItemId);

            parent.UpdateModified();
            destFolder.UpdateModified();
        }

        internal void DeleteThisItem()
        {
            DeleteThisItem(GetParent());
        }

        internal void DeleteThisItem(DavFolder parentFolder)
        {
            Context.ExecuteNonQuery(
                "DELETE FROM Item WHERE ItemId = @ItemID",
                "@ItemID", ItemId);

            if (parentFolder != null)
            {
                parentFolder.UpdateModified();
            }
        }
        
        //!<# if (Params.Class2) { #>
        private List<LockInfo> GetLocks(Guid itemId, bool onlyDeep)
        {
            if (onlyDeep)
            {
                string command = 
                    @"SELECT Token, Shared, Deep, Expires, Owner
                      FROM Lock
                      WHERE ItemID = @ItemID AND Deep = @Deep AND (Expires IS NULL OR Expires > GetUtcDate())";

                return Context.ExecuteLockInfo(
                    command,
                    "@ItemID", itemId,
                    "@Deep", true);
            }

            string selectCommand = 
               @"SELECT Token, Shared, Deep, Expires, Owner
                 FROM Lock
                 WHERE ItemID = @ItemID AND (Expires IS NULL OR Expires > GetUtcDate())";

            return Context.ExecuteLockInfo(
                selectCommand,
                "@ItemID", itemId);
        }

        internal bool ClientHasToken
        {
            get
            {
                List<LockInfo> itemLocks = GetActiveLocks().ToList();
                if (itemLocks.Count == 0)
                {
                    return true;
                }

                IList<string> clientLockTokens = Context.Request.ClientLockTokens;
                return itemLocks.Where(il => clientLockTokens.Contains(il.Token)).Any();
            }
        }

        protected bool ItemHasLock(bool skipShared)
        {
            List<LockInfo> locks = GetActiveLocks().ToList();
            if (locks.Count == 0)
            {
                return false;
            }

            return !skipShared || !locks.All(l => l.Level == LockLevel.Shared);
        }

        protected static void FindLocksDown(IHierarchyItem root, bool skipShared)
        {
            IFolder folder = root as IFolder;
            if (folder != null)
            {
                foreach (IHierarchyItem child in folder.GetChildren(new PropertyName[0]))
                {
                    DavHierarchyItem dbchild = child as DavHierarchyItem;
                    if (dbchild.ItemHasLock(skipShared))
                    {
                        MultistatusException mex = new MultistatusException();
                        mex.AddInnerException(dbchild.Path, new LockedException());
                        throw mex;
                    }

                    FindLocksDown(child, skipShared);
                }
            }
        }
        //!<# } #>

        internal void UpdateModified()
        {
            Context.ExecuteNonQuery(
                "UPDATE Item SET Modified = GETUTCDATE() WHERE ItemId = @ItemId",
                "@ItemId", ItemId);
        }

        protected string CurrentUserName
        {
            get { return Context.User != null ? Context.User.Identity.Name : string.Empty; }
        }

        protected void SetDbField<T>(string columnName, T value)
        {
            string commandText = string.Format("UPDATE Item SET {0} = @value WHERE ItemId = @ItemId", columnName);
            Context.ExecuteNonQuery(
                commandText,
                "@value", value,
                "@ItemId", ItemId);
        }

        //!<# if (Params.MicrosoftExtensions) { #>
        public FileAttributes GetFileAttributes()
        {
            //!<# if (Params.MarkMacOsFilesHidden) { #>
            if (Name.StartsWith("."))
            {
                return fileAttributes | FileAttributes.Hidden;
            }
            //!<# } #>
 
            return fileAttributes;
        }

        public void SetFileAttributes(FileAttributes value)
        {
            SetDbField("FileAttributes", (int)value);
        }
        //!<# } #>
    }
}
